/*
 * ijs.cpp
 * Authors: Soren Johnson and Leonid Reyzin (reyzin@cs.bu.edu)
 *
 * The algorithms in this file implement the so-called
 * Improved Juels-Sudan Secure Sketch, Construction 5 of
 * "Fuzzy Extractors: How to Generate Strong Keys
 * from Biometrics and Other Noisy Data" by
 * Dodis, Ostrovsky, Reyzin, Smith (SIAM Journal on Computing,
 * 38(1):87-139, 2008), http://arxiv.org/abs/cs/0602007
 *
 * See ijs.txt for more information.  This code and explanatory notes
 * are hosted at http://www.cs.bu.edu/~reyzin/code/fuzzy.html
 *
 * Uses Victor Shoup's NTL, see http://www.shoup.net 
 *
 */

#include "ijs.h"
static bool CheckIfDistinctRoots(const ZZ_pX & f);


/*
 * Produces the secure sketch.
 *
 * Namely, calculates the characteristic polynomial of the input set 
 * of size s and outputs the coefficients at degree s-1, ..., s-t as
 * the sketch of the set.  
 *
 * [The characteristic polynomial of a set
 * w = {x_1, x_2, ... x_s} is defined as
 * the  monic polynomial with roots that are
 * the elements of w, i.e.:  (x - x_1) * (x - x_2) * ... * (x - x_s)]
 * 
 * The resulting sketch is placed into the first argument
 *
 * Assumes (does not check) that ZZ_p::init has been run, ZZ_p::modulus()
 * is a prime, t is even, the set contains no element more that once,
 * and that t<=s.
 */
void IJS_Sketch(vec_ZZ_p & sketch, const vec_ZZ_p & set, int t) 
{
  /*************** NOTE ******************************
   This is not the most efficient way to compute the set
   if the length s of the set is much bigger than t,
   because only the top t coefficients are needed.
   By seaching the literature in computational number theory,
   and/or asking experts, we should be able to find a more
   efficient algorithm
  */

  // Build the characteristic polynomial
  ZZ_pX p;
  BuildFromRoots(p, set);

  // the sketch is the t coefficients from degrees s-1 to s-t
  // (i.e., t+1 highest except the leading 1)
  sketch.SetLength (t);
  int s = set.length();
  for (int i = 1; i <= t; i++)
    sketch[i-1] = coeff(p, s - i);
}


/* 
 * Let t = sketch.length() and s = receivedSet.length(). 
 * This function recovers the original set given its sketch
 * and the received set of distance at most t.
 *
 * More formally, this function finds, if it exists,
 * the unique set of size s that has the following two properties:
 * (a) Its symmetric difference with receivedSet is at most t
 * (b) It has the given sketch
 *
 * Returns true and puts the result into the argument recoveredSet
 * if such a set exists.  Returns false if it doesn't.
 * 
 * Assumes (does not check) that ZZ_p::init has been run, that
 * ZZ_p::modulus() is a prime, that t is even, and that receivedSet
 * contains no element more than once, and that t<=s
 */
bool IJS_Recover(vec_ZZ_p & recoveredSet, const vec_ZZ_p & receivedSet, const vec_ZZ_p & sketch)
{
  /*************** NOTE ******************************
   This is not the most efficient way to recover the
   set.  First, there are faster Reed-Solomon 
   decoding algorithms than Welch-Berlekamp, which takes O(s^3) time
   (see comments in the WelchBerlekamp code).  Second, the
   root-finding step at the end may be sped up because
   we know many of the roots are in the receivedSet and 
   we also know which ones are not roots, since they are
   the errors that are found by the error-locating polynomial E
   in the Welch-Berlekamp decoding.  Finding a more
   efficient implementation may be a nice problem.
  */
  int s = receivedSet.length (); // the size of the received, and, therefore, recovered, set
  int t = sketch.length(); // the maximum size of the symmetric difference

  // pHigh is the polynomial of degree s whose top coefficient is 1,
  // the next t coefficients come from the secure sketch,
  // and the remaining coefficients are 0
  ZZ_pX pHigh(s, 1); // initialize to X^s
  for (int i = 1; i <= t; i++)
    SetCoeff(pHigh, (s - i), sketch[i-1]);

  // evaluate pHigh(x) for every x in receivedSet
  // put the result in a vector called y
  vec_ZZ_p y;
  eval (y, pHigh, receivedSet);


  // Use the Welch-Berlekamp algorithm for
  // Reed-Solomon Decoding to find a polynomial pLow of degree
  // at most s-t-1 such that pLow(x[i])=y[i] for at least s-t/2 values
  // of i
  ZZ_pX pLow;
  if(!WelchBerlekamp (pLow, receivedSet, y, t/2))
    return false;

  pHigh -= pLow; // Now pHigh the characterstic polynomial of the recoveredSet
  // before we call FindRoots, we have to check that the polynomial splits into roots
  // since otherwise FindRoots will hang indefinitely (it is possible that, when
  // distance between sets is bigger than t, WelchBerlekamp finds a pLow, 
  // but pHigh-pLow doesn't split into roots -- e.g., t=4, n=13, original set 
  // [0 1 2 3 4 5 6 7], received set [0 1 2 3 4 8 11 12]).
  if (!CheckIfDistinctRoots(pHigh))
    return false;
  FindRoots (recoveredSet, pHigh); // Its roots are the set
  return true;
}



/* 
 * Returns true if f fully factors into distinct roots (i.e., if f is a
 * product of distinct monic degree-1 polynomials times possibly a
 * constant) and false otherwise.  If f is zero, returns false.  Assumes
 * (does not check) that ZZ_p::modulus is a prime
 *
 *  ALGORITHM:
 *  Let n=ZZ_p::modulus() (i.e., the field is GF(n)).
 *  The check is accomplished by checking if f divides X^n - X,
 *  or equivalently if X^n-X is 0 mod f.
 *  X^n - X has n distinct roots by Fermat's little theorem -- namely,
 *  every element of the field is a root.  Hence, f divides it if and only
 *  if f has all its roots and they are all distinct.
 *
 */
static
bool CheckIfDistinctRoots(const ZZ_pX & f)
{
        if (IsZero(f))
          return false;
        // We hanlde degree 0 and degree 1 case separately, so that later
        // we can assume X mod f is the same as X
        if (deg(f) == 0 || deg(f) == 1)
          return true;

        ZZ_pXModulus F(f); // F is the same as f, just more efficient modular operations

        ZZ_pX h;
	ZZ_pX X;
	SetX (X);
        PowerMod (h, X, ZZ_p::modulus(),  F); // h = X^n mod F

        // If X^n - X = 0 mod F, then X^n mod F
        // should be just X (because degree of F > 1)
        return (IsX(h));
}
